import argparse
import re
import sys
from pathlib import Path

from rich import box
from rich.prompt import Confirm
from rich.table import Table

from ._checks import pre_run_check
from ._version import __version__
from .console import console
from .read_filter import get_filter_count


def get_args():
    parser = argparse.ArgumentParser(
        prog="pycashier",
        usage="%(prog)s [-h] sourcedir"
        #       formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    parser.add_argument(
        "sourcedir", help="directory containing sam/fastq files to process"
    )
    parser.add_argument(
        "-t",
        "--threads",
        help="number of cpu cores to use (default: %(default)s)",
        metavar="",
        default=1,
    )
    parser.add_argument(
        "-sc",
        "--single_cell",
        help="turn unampped sam files into cell barcode & umi labeled tsv",
        action="store_true",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="show output of command line calls",
        action="store_true",
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="output directory (default: %(default)s)",
        metavar="",
        default="outs",
    )
    parser.add_argument(
        "-pd",
        "--pipelinedir",
        help="directory containing pipeline \
            output files (default: %(default)s)",
        metavar="",
        default="pipeline",
    )
    parser.add_argument(
        "--skip-init-check",
        help=argparse.SUPPRESS,
        action="store_true",
    )
    parser.add_argument(
        "-V",
        "--version",
        help="show version information and exit",
        action="version",
        version="%(prog)s " + __version__,
    )

    # extract specific parameters
    extract_parser = parser.add_argument_group(title="extract options")

    extract_parser.add_argument(
        "-e",
        "--error",
        help="error tolerance supplied to cutadapt (default: %(default)s)",
        metavar="",
        default=0.1,
        type=float,
    )
    extract_parser.add_argument(
        "-bl",
        "--barcode_length",
        help="integer length of expected barcode (default: %(default)s)",
        metavar="",
        type=int,
        default=20,
    )
    extract_parser.add_argument(
        "-mbl",
        "--min_barcode_length",
        help="minimum length of expected barcode (default: %(default)s)",
        metavar="",
        type=int,
        default=20,
    )
    extract_parser.add_argument(
        "-ua",
        "--upstream_adapter",
        help="5' sequence flanking barcode",
        metavar="",
        type=str,
        default="ATCTTGTGGAAAGGACGAAACACCG",
    )
    extract_parser.add_argument(
        "-da",
        "--downstream_adapter",
        help="3' sequence flanking barcode region",
        metavar="",
        type=str,
        default="GTTTTAGAGCTAGAAATAGCAAGTT",
    )
    extract_parser.add_argument(
        "-ul",
        "--unlinked_adapters",
        help="run cutadapt using unlinked adapters",
        action="store_true",
    )
    extract_parser.add_argument(
        "-q",
        "--quality",
        help="minimum PHRED quality to use to \
            filter reads (default: %(default)s)",
        metavar="",
        default=30,
        type=int,
    )

    cluster_parser = parser.add_argument_group(title="cluster options")
    cluster_parser.add_argument(
        "-r",
        "--ratio",
        help="ratio to use for message passing clustering (default: %(default)s)",
        metavar="",
        default=3,
        type=int,
    )
    cluster_parser.add_argument(
        "-d",
        "--distance",
        help="levenshtein distance for clustering (default: %(default)s)",
        metavar="",
        default=1,
        type=int,
    )

    filter_parser = parser.add_argument_group("filter_options")
    filter_parser.add_argument(
        "-fc",
        "--filter_count",
        help="nominal number of reads for sequence to pass filter",
        metavar="",
        default=None,
        type=int,
    )
    filter_parser.add_argument(
        "-fp",
        "--filter_percent",
        help="minimum percentage of total reads for sequence to pass filter",
        metavar="",
        default=0.005,
        type=float,
    )

    merge_parser = parser.add_argument_group(title="merge options")
    merge_parser.add_argument(
        "-m",
        "--merge",
        help="merge R1 and R2 fastq files using pear",
        action="store_true",
    )
    merge_parser.add_argument(
        "-ko",
        "--keep_output",
        help="keep auxiliary files output by pear in mergefastqs directory",
        action="store_true",
    )
    merge_parser.add_argument(
        "-pa",
        "--pear_args",
        help="additional arguments to pass to pear",
        metavar="",
        type=str,
        default="",
    )

    args = parser.parse_args()

    if not args.skip_init_check:
        pre_run_check()

    return {
        "main": {
            "sourcedir": args.sourcedir,
            "threads": args.threads,
            "quality": args.quality,
            "verbose": args.verbose,
            "outdir": args.outdir,
            "pipelinedir": args.pipelinedir,
        },
        "extract": {
            "error_rate": args.error,
            "upstream_adapter": args.upstream_adapter,
            "downstream_adapter": args.downstream_adapter,
            "barcode_length": args.barcode_length,
            "min_barcode_length": args.min_barcode_length,
            "unlinked_adapters": args.unlinked_adapters,
        },
        "cluster": {
            "ratio": args.ratio,
            "distance": args.distance,
        },
        "merge": {
            "merge": args.merge,
            "keep_output": args.keep_output,
            "pear_args": args.pear_args,
        },
        "filter": {
            "filter_percent": args.filter_percent,
            "filter_count": args.filter_count,
        },
        "single_cell": args.single_cell,
    }


def make_sample_check_table(samples, args):
    processed_samples = []
    table = Table(
        title="[yellow]Samples Queued For Processing",
        box=box.HORIZONTALS,
        header_style="bold bright_blue",
        min_width=80,
    )
    table.add_column(
        "Sample",
        justify="center",
        style="green",
        no_wrap=True,
    )
    table.add_column(
        f"Read Quality\n(Phred Score)\n {args['quality']} ", justify="center"
    )
    table.add_column(
        f"Clustering\n(ratio, distance)\n {args['ratio']}, {args['distance']}",
        justify="center",
    )
    if args["filter_count"] is not None:
        table.add_column(
            f"Filter Cutoff\n(min reads)\n {args['filter_count']}", justify="center"
        )
    else:
        table.add_column(
            f"Filter Cutoff\n(min %)\n {args['filter_percent']} %", justify="center"
        )
    table.add_column("Processed?", justify="center")

    for sample in samples:
        row, processed = check_pipeline_outs(sample, args)

        if processed:
            processed_samples.append(sample)

        table.add_row(*row)

    console.print(table, justify="center")

    print()

    return processed_samples


def check_pipeline_outs(sample, args):
    pipeline = Path(args["pipelinedir"])
    row_list = [sample]

    p1 = re.compile(
        fr'{sample}\.barcodes\.q{args["quality"]}.\
            r{args["ratio"]}d{args["distance"]}\.tsv'
    )
    p2 = re.compile(fr'{sample}\.barcodes\.q{args["quality"]}\.tsv')
    p3 = re.compile(
        fr'{sample}\.barcodes\.q{args["quality"]}.\
            r{args["ratio"]}d{args["distance"]}\.min(?P<filter_count>\d+)\.tsv'
    )

    filters = []
    for f in pipeline.iterdir():
        m = p1.search(f.name)

        if m:
            row_list += ["[bold green]\u2713"] * 2

            filters = []
            for f2 in sorted([f.name for f in Path(args["outdir"]).iterdir()]):
                m = p3.search(f2)
                if m:
                    if args["filter_count"] is not None:
                        filter_count_check = args["filter_count"]
                    else:
                        filter_count_check = get_filter_count(f, args["filter_percent"])

                    if str(filter_count_check) == m.group("filter_count"):
                        filters.append(f"[green]{m.group('filter_count')}[/green]")
                    else:
                        filters.append(m.group("filter_count"))

            if filters:
                row_list.append(", ".join(filters))

            break

    if row_list == [sample]:
        for f in pipeline.iterdir():
            m = p2.search(f.name)
            if m:
                row_list += [":heavy_check_mark:"]

    row_list.extend(["[yellow]Queued"] * (4 - len(row_list)))

    if r"[green]" in "".join(filters):
        row_list.append("[bold green]\u2713")
        processed = True
    else:
        row_list.append("[bold red] Incomplete")
        processed = False

    return row_list, processed


def sample_check(sourcedir, fastqs, cli_args):

    args = {
        "quality": cli_args["main"]["quality"],
        "ratio": cli_args["cluster"]["ratio"],
        "distance": cli_args["cluster"]["distance"],
        "filter_percent": cli_args["filter"]["filter_percent"],
        "filter_count": cli_args["filter"]["filter_count"],
        "pipelinedir": cli_args["main"]["pipelinedir"],
        "outdir": cli_args["main"]["outdir"],
    }

    samples = [f.name.split(".")[0] for f in fastqs]
    # outs_files = sorted([f.name for f in Path(args['outdir']).iterdir()])
    # check_outs(samples, outs_files)\
    processed_samples = make_sample_check_table(samples, args)

    if not Confirm.ask("Continue with these samples?"):
        sys.exit()

    return processed_samples


####################################################
#  functions to get additional samples found in outs
####################################################

# def get_params(f):
#     p = re.compile(
#         r'(?P<sample>\w+)\.barcodes\.q(?P<quality>\d\d).r(?P<ratio>\d+)d(?P<distance>\d+)\.min(?P<filter_count>\d+)\.tsv'
#     )
#     m = p.search(f)
#     if m:
#         return m.groupdict()
#     else:
#         raise ValueError(f"Unexpected file in outs directory:{f}")

# def check_outs(samples, files):
#     table_rows = []

#     for f in files:
#         params = get_params(f)
#         if params['sample'] not in samples:
#             table_rows.append([
#                 params['sample'], params['quality'],
#                 f"{params['ratio']}, {params['distance']}",
#                 params['filter_count']
#             ])
#     if table_rows:
#         console = Console()
#         table = Table(title="Additional Samples Found in Outs Directory",
#                       box=box.HORIZONTALS,
#                       header_style="bold bright_blue",
#                       min_width=70)
#         table.add_column("Sample",
#                          justify="center",
#                          style="green",
#                          no_wrap=True)
#         table.add_column("Read Quality\n(Phred Score)", justify="center")
#         table.add_column('Clustering\n(ratio,distance)', justify='center')
#         table.add_column('Filter Cutoff\n(min reads)', justify='center')
#         for row in table_rows:
#             table.add_row(*row)
#         console.print(table)



import shlex
import subprocess
import sys
import tempfile
from pathlib import Path

try:
    import pysam
except ImportError:
    pass

from rich.progress import Progress

from .console import console


def sam_to_name_labeled_fastq(sample, in_file, out_file):

    new_sam = False

    with open(in_file, "r") as f_in:
        if f_in.readline()[0:3] != "@HD":
            new_sam = True

    if new_sam:
        console.log(f"[green]{sample}[/green]: sam is headerless, adding a fake one")
        sam_file = fake_header_add(in_file)
    else:
        sam_file = in_file

    sam_length = pysam.AlignmentFile(sam_file, "r", check_sq=False).count()
    sam = pysam.AlignmentFile(sam_file, "r", check_sq=False)

    with open(out_file, "w") as f_out:

        with Progress() as progress:
            task = progress.add_task("[red] Converting sam to fastq", total=sam_length)

            for record in sam:

                tagdict = dict(record.tags)
                cell_barcode = None
                if "CB" in tagdict.keys():
                    cell_barcode = tagdict["CB"].split("-")[0]
                elif "CR" in tagdict.keys():
                    cell_barcode = tagdict["CR"]

                umi = None
                if "UB" in tagdict.keys():
                    umi = tagdict["UB"]
                elif "UR" in tagdict.keys():
                    umi = tagdict["UR"]

                if cell_barcode and umi:

                    qualities = record.query_qualities
                    ascii_qualities = "".join([chr(q + 33) for q in qualities])

                    f_out.write(f"@{record.query_name}_{umi}_{cell_barcode}\n")
                    f_out.write(f"{record.query_sequence}\n+\n{ascii_qualities}\n")

                progress.advance(task)

    if new_sam:

        Path(sam_file).unlink()


def labeled_fastq_to_tsv(in_file, out_file):

    with open(in_file) as f_in:

        with open(out_file, "w") as f_out:
            read_lines = []
            # TODO: add progress bar
            for line in f_in.readlines():
                read_lines.append(line)
                if len(read_lines) == 4:

                    read_name, umi, cell_barcode = read_lines[0].rstrip("\n").split("_")
                    lineage_barcode = read_lines[1].rstrip("\n")

                    f_out.write(
                        f"{read_name}\t{umi}\
                            \t{cell_barcode}\t{lineage_barcode}\n"
                    )
                    read_lines = []


def fake_header_add(in_file):

    fake_header = """@HD\tVN:1.6\tSO:coordinate
@SQ\tSN:1\tLN:1000
@SQ\tSN:2\tLN:1000
@SQ\tSN:3\tLN:1000
@SQ\tSN:4\tLN:1000
@SQ\tSN:5\tLN:1000
@SQ\tSN:6\tLN:1000
@SQ\tSN:7\tLN:1000
@SQ\tSN:8\tLN:1000
@SQ\tSN:9\tLN:1000
@SQ\tSN:10\tLN:1000
@SQ\tSN:11\tLN:1000
@SQ\tSN:12\tLN:1000
@SQ\tSN:13\tLN:1000
@SQ\tSN:14\tLN:1000
@SQ\tSN:15\tLN:1000
@SQ\tSN:16\tLN:1000
@SQ\tSN:17\tLN:1000
@SQ\tSN:18\tLN:1000
@SQ\tSN:19\tLN:1000
@SQ\tSN:20\tLN:1000
@SQ\tSN:21\tLN:1000
@SQ\tSN:22\tLN:1000
@SQ\tSN:23\tLN:1000
@SQ\tSN:X\tLN:1000
@SQ\tSN:Y\tLN:1000
@SQ\tSN:MT\tLN:1000
@SQ\tSN:GL000192.1\tLN:1000
@SQ\tSN:GL000225.1\tLN:1000
@SQ\tSN:GL000194.1\tLN:1000
@SQ\tSN:GL000193.1\tLN:1000
@SQ\tSN:GL000200.1\tLN:1000
@SQ\tSN:GL000222.1\tLN:1000
@SQ\tSN:GL000212.1\tLN:1000
@SQ\tSN:GL000195.1\tLN:1000
@SQ\tSN:GL000223.1\tLN:1000
@SQ\tSN:GL000224.1\tLN:1000
@SQ\tSN:GL000219.1\tLN:1000
@SQ\tSN:GL000205.1\tLN:1000
@SQ\tSN:GL000215.1\tLN:1000
@SQ\tSN:GL000216.1\tLN:1000
@SQ\tSN:GL000217.1\tLN:1000
@SQ\tSN:GL000199.1\tLN:1000
@SQ\tSN:GL000211.1\tLN:1000
@SQ\tSN:GL000213.1\tLN:1000
@SQ\tSN:GL000220.1\tLN:1000
@SQ\tSN:GL000218.1\tLN:1000
@SQ\tSN:GL000209.1\tLN:1000
@SQ\tSN:GL000221.1\tLN:1000
@SQ\tSN:GL000214.1\tLN:1000
@SQ\tSN:GL000228.1\tLN:1000
@SQ\tSN:GL000227.1\tLN:1000
@SQ\tSN:GL000191.1\tLN:1000
@SQ\tSN:GL000208.1\tLN:1000
@SQ\tSN:GL000198.1\tLN:1000
@SQ\tSN:GL000204.1\tLN:1000
@SQ\tSN:GL000233.1\tLN:1000
@SQ\tSN:GL000237.1\tLN:1000
@SQ\tSN:GL000230.1\tLN:1000
@SQ\tSN:GL000242.1\tLN:1000
@SQ\tSN:GL000243.1\tLN:1000
@SQ\tSN:GL000241.1\tLN:1000
@SQ\tSN:GL000236.1\tLN:1000
@SQ\tSN:GL000240.1\tLN:1000
@SQ\tSN:GL000206.1\tLN:1000
@SQ\tSN:GL000232.1\tLN:1000
@SQ\tSN:GL000234.1\tLN:1000
@SQ\tSN:GL000202.1\tLN:1000
@SQ\tSN:GL000238.1\tLN:1000
@SQ\tSN:GL000244.1\tLN:1000
@SQ\tSN:GL000248.1\tLN:1000
@SQ\tSN:GL000196.1\tLN:1000
@SQ\tSN:GL000249.1\tLN:1000
@SQ\tSN:GL000246.1\tLN:1000
@SQ\tSN:GL000203.1\tLN:1000
@SQ\tSN:GL000197.1\tLN:1000
@SQ\tSN:GL000245.1\tLN:1000
@SQ\tSN:GL000247.1\tLN:1000
@SQ\tSN:GL000201.1\tLN:1000
@SQ\tSN:GL000235.1\tLN:1000
@SQ\tSN:GL000239.1\tLN:1000
@SQ\tSN:GL000210.1\tLN:1000
@SQ\tSN:GL000231.1\tLN:1000
@SQ\tSN:GL000229.1\tLN:1000
@SQ\tSN:GL000226.1\tLN:1000
"""
    # TODO: change to with statement
    f = tempfile.NamedTemporaryFile(delete=False, dir=Path.cwd())
    # print(f'copying new tmp sam to {f.name}')
    f.write((bytes(fake_header, encoding="utf-8")))
    f.flush()

    subprocess.run(["cat", in_file], stdout=f)

    # #python implementation
    # with open(in_file, 'r') as sam_file:
    #     for line in sam_file:
    #         f.write(bytes(line, encoding='utf-8'))
    #    #f.write(bytes(sam_file.read()), encoding='utf-8')

    f.close()
    return f.name


def single_cell_process(sample, f, sourcedir, cli_args, status):

    error_rate = cli_args["extract"]["error_rate"]
    threads = cli_args["main"]["threads"]
    barcode_length_min = 10
    barcode_length = cli_args["extract"]["barcode_length"]
    upstream_adapter = cli_args["extract"]["upstream_adapter"]
    downstream_adapter = cli_args["extract"]["downstream_adapter"]
    adapter_string = f"-g {upstream_adapter} -a {downstream_adapter}"

    input_file = f
    pipeline_dir = Path(cli_args["main"]["pipelinedir"])
    fastq_out = pipeline_dir / f"{sample}.cell_record_labeled.fastq"
    output_file = pipeline_dir / f"{sample}.cell_record_labeled.barcode.fastq"
    tsv_out = (
        Path(cli_args["main"]["outdir"]) / f"{sample}.cell_record_labeled.barcode.tsv"
    )

    if not fastq_out.is_file():
        console.log(f"[green]{sample}[/green]: converting sam to labeled fastq")
        status.stop()
        sam_to_name_labeled_fastq(sample, input_file, fastq_out)
        status.start()
    else:
        console.log(
            f"[green]{sample}[/green]: skipping sam to labeled fastq conversion"
        )

    if not output_file.is_file():

        console.log(f"[green]{sample}[/green]: extracting barcodes")

        command = f"cutadapt \
            -e {error_rate} \
            -j {threads} \
            --minimum-length={barcode_length_min} \
            --maximum-length={barcode_length} \
            --max-n=0 \
            --trimmed-only {adapter_string} \
            -n 2 \
            -o {output_file} {fastq_out}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if cli_args["main"]["verbose"]:
            console.print("[yellow]CUTADAPT OUTPUT:")
            console.print(p.stdout)

    if not tsv_out.is_file():
        console.log(f"[green]{sample}[/green]: converting labeled fastq to tsv")
        labeled_fastq_to_tsv(output_file, tsv_out)

    else:
        console.log(
            f"[green]{sample}[/green]: skipping labeled fastq to tsv conversion"
        )


def single_cell(sourcedir, cli_args):

    console.rule("SINGLE CELL MODE", align="center", style="red")
    print()

    sam_files = [f for f in sourcedir.iterdir()]

    for f in sam_files:

        ext = f.suffix

        if ext != ".sam":
            raise ValueError("There is a non sam file in the provided input directory:")

    for f in sam_files:

        sample = f.name.split(".")[0]

        with console.status(
            f"Processing sample: [green]{sample}[/green]", spinner="dots12"
        ) as status:
            single_cell_process(sample, f, sourcedir, cli_args, status)

        console.log(f"[green]{sample}[/green]: processing completed")
        console.rule()

    console.print("\n[green]FINISHED!")
    sys.exit()



import sys
from shutil import which

from rich.table import Table

from .console import console

PACKAGES = ["cutadapt", "fastq_quality_filter", "pysam", "starcode"]


def pre_run_check():

    pkgs_exist = {name: is_tool(name) for name in PACKAGES}
    if False in pkgs_exist.values():
        console.print("\n[red bold] FAILED PRE-RUN CHECKS!!\n")
        table = Table(title="Pre-run Dependency Check")
        table.add_column("package", justify="right")
        table.add_column("found?", justify="left")
        for name, exists in pkgs_exist.items():
            if exists:
                found = "[green] yes"
            else:
                found = "[red] no"
            table.add_row(name, found)
        console.print(table)
        console.print(
            "It's recommeneded to install pycashier within a conda environment."
        )
        console.print(
            "See README on github for details: [link]https://github.com/brocklab/pycashier[/link]"
        )
        sys.exit(1)


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""
    if name == "pysam":
        try:
            import pysam  # noqa

            return True
        except ImportError:
            return False
    return which(name) is not None



import csv
from pathlib import Path

from .console import console


def get_filter_count(file_in, filter_percent):
    total_reads = 0

    with open(file_in, newline="") as csvfile:
        spamreader = csv.reader(csvfile, delimiter="\t")
        for row in spamreader:
            total_reads += float(row[1])

    filter_count = int(round(total_reads * filter_percent / 100, 0))

    return filter_count


def filter_by_percent(file_in, filter_percent, outdir):

    filter_by_count(file_in, get_filter_count(file_in, filter_percent), outdir)


def filter_by_count(file_in, filter_count, outdir):

    name = file_in.stem
    ext = file_in.suffix
    csv_out = Path(outdir) / f"{name}.min{filter_count}{ext}"

    with open(file_in, "r") as csv_in:
        with open(csv_out, "w") as csv_out:
            for line in csv_in:
                linesplit = line.split("\t")
                if int(linesplit[1]) >= filter_count:
                    csv_out.write(f"{linesplit[0]}\t{linesplit[1]}")


def read_filter(
    sample, filter_count, filter_percent, quality, ratio, distance, **kwargs
):
    outdir = kwargs["outdir"]
    file_in = (
        Path(kwargs["pipelinedir"])
        / f"{sample}.barcodes.q{quality}.r{ratio}d{distance}.tsv"
    )

    if filter_count is not None:
        console.log(
            f"[green]{sample}[/green]: removing "
            f"sequences with less than {filter_count} reads"
        )

        filter_by_count(file_in, filter_count, outdir)

    else:
        console.log(
            f"[green]{sample}[/green]: removing"
            f" sequences with less than {filter_percent}% of total reads"
        )

        filter_by_percent(file_in, filter_percent, outdir)



import shlex
import subprocess
from pathlib import Path

from .console import console
from .utils import extract_csv_column


def cluster(sample, ratio, distance, quality, threads, **kwargs):
    pipeline = Path(kwargs["pipelinedir"])

    extracted_csv = pipeline / f"{sample}.barcodes.q{quality}.tsv"

    input_file = extract_csv_column(extracted_csv, 2)

    output_file = pipeline / f"{sample}.barcodes.q{quality}.r{ratio}d{distance}.tsv"

    if not output_file.is_file():
        console.log(f"[green]{sample}[/green]: clustering barcodes")
        command = f"starcode -d {distance} -r {ratio} -t {threads} -i {input_file} -o {output_file}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        if kwargs["verbose"]:
            console.print("[yellow]STARCODE OUTPUT:")
            console.print(p.stdout)

        console.log(f"[green]{sample}[/green]: clustering complete")

    else:
        console.log(f"[green]{sample}[/green]: skipping clustering")



import shlex
import subprocess
import sys
from pathlib import Path

from .console import console
from .utils import fastq_to_csv


def extract(
    sample,
    fastq,
    sourcedir,
    error_rate,
    threads,
    barcode_length,
    min_barcode_length,
    upstream_adapter,
    downstream_adapter,
    unlinked_adapters,
    quality,
    **kwargs,
):

    pipeline = Path(kwargs["pipelinedir"])
    barcode_fastq = pipeline / f"{sample}.barcode.fastq"
    input_file = fastq
    filtered_barcode_fastq = pipeline / f"{sample}.barcode.q{quality}.fastq"

    if unlinked_adapters:
        adapter_string = f"-g {upstream_adapter} -a {downstream_adapter}"
    else:
        adapter_string = f"-g {upstream_adapter}...{downstream_adapter}"

    if not filtered_barcode_fastq.is_file():

        console.log(f"[green]{sample}[/green]: extracting and filtering barcodes")

        command = f"cutadapt \
            -e {error_rate} \
            -j {threads} \
            --minimum-length={min_barcode_length} \
            --maximum-length={barcode_length} \
            --max-n=0 \
            --trimmed-only {adapter_string} \
            -n 2 \
            -o {barcode_fastq} {input_file}"

        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if barcode_fastq.stat().st_size == 0:
            console.print("[yellow]CUTADAPT OUTPUT:")
            console.print(p.stdout)
            console.print(
                f"[green]{sample}[/green]: Failed to extract reads for sample"
            )
            console.print("see above for cutadapt output")
            sys.exit()

        elif kwargs["verbose"]:
            console.print("[yellow]CUTADAPT OUTPUT:")
            console.print(p.stdout)

        command = f"fastq_quality_filter \
            -q {quality} \
            -p 100 \
            -Q 33 \
            -i {barcode_fastq} \
            -o {filtered_barcode_fastq}"

        args = shlex.split(command)

        p = subprocess.run(args)

        barcode_fastq.unlink()

        console.log(f"[green]{sample}[/green]: extraction and filtering complete")

    else:
        console.log(f"[green]{sample}[/green]: skipping extraction")

    barcodes_out = pipeline / f"{sample}.barcodes.q{quality}.tsv"

    if not barcodes_out.is_file():
        console.log(f"[green]{sample}[/green]: converting fastq to tsv")
        fastq_to_csv(filtered_barcode_fastq, barcodes_out)
    else:
        console.log(f"[green]{sample}[/green]: skipping fastq to tsv conversion")



import re
import shlex
import subprocess
import sys
from pathlib import Path

from .console import console


def merge_single(
    sample, fastqs, sourcedir, threads, verbose, pipelinedir, pear_args, keep_output
):
    pipeline = Path(pipelinedir)

    # TODO: refactor for clarity and memory usage
    for f in fastqs:

        R1_regex = r"" + re.escape(sample) + r"\..*R1.*\.fastq\.gz"
        m = re.search(R1_regex, f.name)
        if m:
            R1_file = m.group(0)

        R2_regex = r"" + re.escape(sample) + r"\..*R2.*\.fastq\.gz"
        m = re.search(R2_regex, f.name)
        if m:
            R2_file = m.group(0)

    if R1_file is None or R2_file is None:
        print("oops I didnt find an R1 or R2 file")
        sys.exit(1)

    mergedfastq = Path("mergedfastqs")
    merged_barcode_fastq = mergedfastq / f"{sample}.merged.raw.fastq"
    merged_barcode_file_prefix = pipeline / f"{sample}.merged.raw"

    files = [R1_file, R2_file]

    if not merged_barcode_fastq.is_file():

        console.log(f"[green]{sample}[/green]: extracting and moving fastqs")
        # future implementations may use a python based extraction (using gzip)
        # TODO: Make fastq extraction conditional

        path_to_r1 = sourcedir / R1_file
        path_to_r2 = sourcedir / R2_file

        command = f"gunzip -k {path_to_r1} {path_to_r2}"
        args = shlex.split(command)
        p = subprocess.run(args)

        files = [Path(f) for f in files]  # replace with sample dict of files

        for f in files:

            old_path = sourcedir / f.stem
            new_path = pipeline / f.stem
            old_path.rename(new_path)

        console.log(f"[green]{sample}[/green]: starting fastq merge")
        command = f"pear -f {path_to_r1} -r {path_to_r2} \
            -o {merged_barcode_file_prefix} -j {threads} {pear_args}"
        args = shlex.split(command)

        p = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )

        if verbose:
            console.print("[yellow]PEAR OUTPUT:")
            console.print(p.stdout)

        # remove the extra files made from pear
        if not keep_output:
            for suffix in [
                "discarded.fastq",
                "unassembled.forward.fastq",
                "unassembled.reverse.fastq",
            ]:
                f = Path(f"{merged_barcode_file_prefix}.{suffix}")
                f.unlink()

        merged_barcode_file_prefix.with_suffix(
            merged_barcode_file_prefix.suffix + ".assembled.fastq"
        ).rename(merged_barcode_fastq)

    else:
        print(f"Found merged barcode fastq for sample:{sample}")
        console.log(f"[green]{sample}[/green]: skipping fastq merge")


def merge(fastqs, sourcedir, cli_args):
    console.rule("MERGE MODE", align="center", style="red")
    print()
    Path("mergedfastqs").mkdir(exist_ok=True)

    samples = []

    for f in fastqs:

        m = re.search(r"(.+?)\..*R.*\.fastq\.gz", f.name)
        if m:
            samples.append(m.group(1))
        else:
            print(f"Failed to obtain sample name from {f}")
            sys.exit(1)

    print("Found the following samples:")
    for s in set(samples):
        console.print(f"[green]{s}")
    print()

    if len(samples) / len(set(samples)) != 2:
        print("There should be an R1 and R2 fastq file for each sample.")
        sys.exit(1)

    for sample in set(samples):

        with console.status(
            f"Processing sample: [green]{sample}[/green]", spinner="dots12"
        ):

            merge_single(
                sample,
                fastqs,
                sourcedir,
                cli_args["main"]["threads"],
                verbose=cli_args["main"]["verbose"],
                pipelinedir=cli_args["main"]["pipelinedir"],
                pear_args=cli_args["merge"]["pear_args"],
                keep_output=cli_args["merge"]["keep_output"],
            )

        console.log(f"[green]{sample}[/green]: processing completed")
        console.rule()

    console.print("\n[green]FINISHED!")
    sys.exit()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycashier']

package_data = \
{'': ['*']}

install_requires = \
['rich>=10.12.0,<11.0.0']

entry_points = \
{'console_scripts': ['pycashier = pycashier.pycashier:main']}

setup_kwargs = {
    'name': 'pycashier',
    'version': '0.2.6',
    'description': 'cash in on expressed barcode tags',
    'long_description': "# Cash in on Expressed Barcode Tags (EBTs) from NGS Sequencing Data *with Python*\n\nTool for extracting and processing DNA barcode tags from Illumina sequencing.\n\nDefault parameters are designed for use by the Brock Lab to process data generated from\nClonMapper lineage tracing experiments, but is extensible to other similarly designed tools.\n\n\n### Bioconda Dependencies\n- cutadapt (sequence extraction)\n- starcode (sequence clustering)\n- fastx-toolkit (PHred score filtering)\n- pear (paired end read merging)\n- pysam (sam file convertion to fastq)\n\n## Installation\nIt's recommended to use [conda](https://docs.conda.io/en/latest/)/[mamba](https://github.com/mamba-org/mamba) to install and manage the dependencies for this package\n\n```bash\nconda install -c conda-forge -c bioconda cutadapt fastx-toolkit pear pysam starcode pycashier\n```\n\nYou can also use the included `environment.yml` to create your environment and install everything you need.\n\n```bash\nconda env create -f https://raw.githubusercontent.com/brocklab/pycashier/main/environment.yml\nconda activate cashierenv\n```\n\nAdditionally you may install with pip. Though it will be up to you to ensure all the\ndependencies you would install from bioconda are on your path and installed correctly.\n`Pycashier` will check for them before running.\n\n```bash\npip install pycashier\n```\n\n## Usage\n\nPycashier has one required argument which is the directory containing the fastq or sam files you wish to process.\n\n```bash\nconda activate cashierenv\npycashier ./fastqs\n```\nFor additional parameters see `pycashier -h`.\n\nAs the files are processed two additional directories will be created `pipeline` and `outs`.\n\n**Note**: these can be specified with `-pd/--pipelinedir` and  `-o/--outdir`.\n\nCurrently all intermediary files generated as a result of the program will be found in `pipeline`.\n\nWhile the final processed files will be found within the `outs` directory.\n\n## Merging Files\n\nPycashier can now take paired end reads and perform a merging of the reads to produce a fastq which can then be used with pycashier's default feature.\n```bash\npycashier ./fastqs -m\n```\n\n## Processing Barcodes from 10X bam files\n\nPycashier can also extract gRNA barcodes along with 10X cell and umi barcodes.\n\nFirstly we are only interested in the unmapped reads. From the cellranger bam output you would obtain these reads using samtools.\n\n```\nsamtools view -f 4 possorted_genome_bam.bam > unmapped.sam\n```\n\nThen similar to normal barcode extraction you can pass a directory of these unmapped sam files to pycashier and extract barcodes. You can also still specify extraction parameters that will be passed to cutadapt as usual.\n\n*Note*: The default parameters passed to cutadapt are unlinked adapters and minimum barcode length of 10 bp.\n\n```\npycashier ./unmapped_sams -sc\n```\nWhen finished the `outs` directory will have a `.tsv` containing the following columns: Illumina Read Info, UMI Barcode, Cell Barcode, gRNA Barcode\n\n\n## Usage notes\nPycashier will **NOT** overwrite intermediary files. If there is an issue in the process, please delete either the pipeline directory or the requisite intermediary files for the sample you wish to reprocess. This will allow the user to place new fastqs within the source directory or a project folder without reprocessing all samples each time.\n- Currently, pycashier expects to find `.fastq.gz` files when merging and `.fastq` files when extracting barcodes. This behavior may change in the future.\n- If there are reads from multiple lanes they should first be concatenated with `cat sample*R1*.fastq.gz > sample.R1.fastq.gz`\n- Naming conventions:\n    - Sample names are extracted from files using the first string delimited with a period. Please take this into account when naming sam or fastq files.\n    - Each processing step will append information to the input file name to indicate changes, again delimited with periods.\n\n\n## Acknowledgments\n\n[Cashier](https://github.com/russelldurrett/cashier) is a tool developed by Russell Durrett for the analysis and extraction of expressed barcode tags.\nThis version like it's predecessor wraps around several command line bioinformatic tools to pull out expressed barcode tags.\n",
    'author': 'Daylin Morgan',
    'author_email': 'daylinmorgan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/brocklab/pycashier/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
